#!/bin/bash -e
#
#  Generate a root CA cert for signing, and then a subject cert.
#  Usage: make-certs.sh hostname [user[@domain]] [more ...]
#  For testing only, probably still has some bugs in it.
#

DOMAIN=redhat.com
DAYS=365

# Cleanup temporary files at exit.
touch openssl.cnf
cleanup() {
	test -f openssl.cnf && rm -f openssl.cnf
	test -f host.csr    && rm -f host.csr
	test -f ca.txt      && rm -f ca.txt
}
trap cleanup EXIT

# Choose a user name part for email attributes.
GIVENUSER=$2
test x$GIVENUSER = x && GIVENUSER=$USER
echo $GIVENUSER | grep -q @ || GIVENUSER=$GIVENUSER@$DOMAIN
DOMAIN=`echo "$GIVENUSER" | cut -f2- -d@`

# Default to generating a key/cert pair for the current host.
commonname="$1"
if test "x$commonname" = x ; then
	echo Usage: `basename $0` 'commonname' user'[@domain]' '[more [...]]'
	echo More:
	echo -e \\tKey usage: "[sign|signing|encrypt|encryption]"
	echo -e \\tAuthority Access Info OCSP responder: "ocsp:URI"
	echo -e \\tCRL distribution point: "crl:URI"
	echo -e \\tSubject Alternative Name:
	echo -e \\t\\tIP address: w.x.y.z
	echo -e \\t\\tEmail address: "*@*.com"
	echo -e \\t\\tKerberos principal name: "*@*.COM"
	echo -e \\tExtended key usage:
	echo -e \\t\\t1....
	echo -e \\t\\t2....
	echo -e \\t\\tid-kp-server-auth \| tls-server
	echo -e \\t\\tid-kp-client-auth \| tls-client
	echo -e \\t\\tid-kp-email-protection \| email
	echo -e \\t\\tid-ms-kp-sc-logon \| id-ms-sc-logon
	echo -e \\t\\tid-pkinit-kp-client-auth \| id-pkinit-client
	echo -e \\t\\tid-pkinit-kp-kdc \| id-pkinit-kdc
	exit 1
fi

shift || true
shift || true

# Done already?
done=:

# Generate a key.
if ! test -f ca.key ; then
umask=`umask -p`
umask 077
openssl genrsa 1024 -nodes > ca.key 2> /dev/null
$umask
else
echo "You already have a ca.key file; not replacing."
done=echo
fi

# Create a new serial number.
if ! test -s ca.srl ; then
(dd if=/dev/urandom bs=8 count=1 2> /dev/null) | od -t x1c | head -n 1 | awk '{$1="00";OFS="";print}' > ca.srl
else
echo "You already have a ca.srl file; not replacing."
fi

# Generate the CA cert.  Set the X.509v3 basic constraints to mark it as a CA
# certificate.
if ! test -f ca.crt ; then
cat > openssl.cnf << EOF
[req]
x509_extensions = v3_ca
distinguished_name = req_ca_name
oid_section = new_oids
prompt = no

[req_ca_name]
commonName = Test Certifying CA
EOF
echo $DOMAIN | awk 'BEGIN {FS="."}{for(i=1;i<=NF;i++){print i-1 ".domainComponent="$i;}}' >> openssl.cnf
cat >> openssl.cnf << EOF

[new_oids]
domainComponent = 0.9.2342.19200300.100.1.25

[v3_ca]
basicConstraints=CA:TRUE
subjectKeyIdentifier=hash
authorityKeyIdentifier=keyid:always,issuer:always
EOF
openssl req -config `pwd`/openssl.cnf -new -key ca.key > ca.csr 2> /dev/null
openssl x509 -extfile `pwd`/openssl.cnf -days $DAYS -req -CAserial ca.srl -signkey ca.key -extensions v3_ca -in ca.csr -out ca.crt
openssl x509 -noout -text -in ca.crt > ca.txt
cat ca.crt >> ca.txt
cat ca.txt > ca.crt
rm ca.txt
else
echo "You already have a ca.crt file; not replacing."
done=echo
fi

# Generate another key.
if ! test -f "$commonname.key" ; then
umask=`umask -p`
umask 077
openssl genrsa 1024 -nodes > "$commonname.key" 2> /dev/null
$umask
else
echo "You already have a $commonname.key file; not replacing."
done=echo
fi

# Generate the subject's certificate.  Set the X.509v3 basic constraints to
# mark it as a non-CA certificate.
if ! test -f "$commonname.crt" ; then
cat > openssl.cnf << EOF
[req]
distinguished_name = req_issued_name
prompt = no

[x509]
x509_extensions = v3_issued

[req_issued_name]
commonName = $commonname
mail = $GIVENUSER
EOF
echo $DOMAIN | awk 'BEGIN {FS="."}{for(i=1;i<=NF;i++){print i-1 ".domainComponent="$i;}}' >> openssl.cnf
cat >> openssl.cnf << EOF

[v3_issued]
basicConstraints=CA:FALSE
subjectKeyIdentifier=hash
authorityKeyIdentifier=keyid,issuer
nsComment="Testing Certificate for $commonname"
EOF

while test $# -gt 0 ; do
	type=
	case "$1" in
	OCSP:*|ocsp:*)
		aiaval=`echo "$1" | cut -f2- -d:`
		aia="OCSP;URI:$aiaval"
		;;
	CRL:*|crl:*)
		crlval=`echo "$1" | cut -f2- -d:`
		crl="URI:$crlval"
		;;
	signing|sign)
		keyusage="${keyusage:+${keyusage},}nonRepudiation,digitalSignature"
		;;
	encryption|encrypt)
		keyusage="${keyusage:+${keyusage},}keyEncipherment"
		;;
	1.*|2.*|id-*|tls-*|email)
		ekuval=`echo "$1" | tr '[A-Z]' '[a-z]' | sed 's,\-,,g'`
		case "$ekuval" in
		idkpserverauth|tlsserver) ekuval=1.3.6.1.5.5.7.3.1;;
		idkpclientauth|tlsclient) ekuval=1.3.6.1.5.5.7.3.2;;
		idkpemailprotection|email) ekuval=1.3.6.1.5.5.7.3.4;;
		idmskpsclogon|idmssclogon) ekuval=1.3.6.1.4.1.311.20.2.2;;
		idpkinitkpclientauth|idpkinitclient) ekuval=1.3.6.1.5.2.3.4;;
		idpkinitkpkdc|idpkinitkdc) ekuval=1.3.6.1.5.2.3.5;;
		esac
		if test -z "$eku" ; then
			eku="$ekuval"
		else
			eku="$eku,$ekuval"
		fi
		;;
	*@*COM)
		type="otherName:1.3.6.1.5.2.2;SEQUENCE:$1,otherName:1.3.6.1.4.1.311.20.2.3;UTF8"
		principals="$principals $1"
		;;
	*@*com|copy)                 type=email;;
	[0-9]*.[0-9]*.[0-9]*.[0-9]*) type=IP;;
	*)                           type=DNS;;
	esac
	if test -n "$type" ; then
		newvalue="${type}:$1"
		if test -z "$altnames" ; then
			altnames="${newvalue}"
		else
			altnames="${altnames},${newvalue}"
		fi
	fi
	shift
done
if test -n "$aia" ; then
	echo "authorityInfoAccess = ${aia}" >> openssl.cnf
fi
if test -n "$crl" ; then
	echo "crlDistributionPoints = ${crl}" >> openssl.cnf
fi
if test -n "$keyusage" ; then
	echo "keyUsage = ${keyusage}" >> openssl.cnf
fi
if test -n "$altnames" ; then
	echo "subjectAltName = ${altnames}" >> openssl.cnf
fi
if test -n "$eku" ; then
	echo "extendedKeyUsage = ${eku}" >> openssl.cnf
fi
echo "subjectKeyIdentifier=hash" >> openssl.cnf
echo "authorityKeyIdentifier=keyid:always,issuer" >> openssl.cnf
for value in $principals; do
	user=`echo "$value" | cut -f1 -d@`
	realm=`echo "$value" | cut -f2- -d@`
	echo "" >> openssl.cnf
	echo "[$value]" >> openssl.cnf
	echo "realm=EXPLICIT:0,GeneralString:$realm" >> openssl.cnf
	echo "kerberosname=EXPLICIT:1,SEQUENCE:krb5$user" >> openssl.cnf

	echo "" >> openssl.cnf
	echo "[krb5$user]" >> openssl.cnf
	echo "nametype=EXPLICIT:0,INTEGER:1" >> openssl.cnf
	echo "namelist=EXPLICIT:1,SEQUENCE:krb5basic$user" >> openssl.cnf

	echo "[krb5basic$user]" >> openssl.cnf
	count=0
	for part in `echo "$user" | sed 's,/, ,g'` ; do
		echo "$count.part=GeneralString:$part" >> openssl.cnf
		count=`expr "$count" + 1`
	done
done

openssl req -config `pwd`/openssl.cnf -new -key "$commonname.key" > host.csr 2> /dev/null
openssl x509 -req -extfile `pwd`/openssl.cnf -extensions v3_issued -CA ca.crt -CAkey ca.key -CAserial ca.srl -in host.csr -days $DAYS > "$commonname.crt" 2> /dev/null
else
echo "You already have a $commonname.crt file; not replacing."
done=echo
fi

# Create ca.pem and the subject's name.pem for the benefit of applications
# which expect both the private key and the certificate in one file.
umask=`umask -p`
umask 077
if ! test -f ca.pem ; then
cat ca.key ca.crt > ca.pem
else
echo "You already have a ca.pem file; not replacing."
done=echo
fi
if ! test -f "$commonname.pem" ; then
cat "$commonname.key" "$commonname.crt" > "$commonname.pem"
else
echo "You already have a $commonname.pem file; not replacing."
done=echo
fi
if ! test -f "$commonname.p12" ; then
openssl pkcs12 -export -inkey "$commonname.key" -in "$commonname.crt" -name "$commonname" -out "$commonname".p12 -nodes -passout pass:
else
echo "You already have a $commonname.p12 file; not replacing."
done=echo
fi
$umask
$done

echo CA certificate:
openssl x509 -noout -issuer  -in ca.crt | sed s,=\ ,\ ,g
openssl x509 -noout -subject -in ca.crt | sed s,=\ ,\ ,g
echo
echo End entity certificate:
openssl x509 -noout -issuer  -in "$commonname.crt" | sed s,=\ ,\ ,g
openssl x509 -noout -subject -in "$commonname.crt" | sed s,=\ ,\ ,g
openssl x509 -noout -serial  -in "$commonname.crt" | sed s,=,\ ,g
echo
echo PKCS12 bag:
openssl pkcs12 -in "$commonname".p12 -nodes -nokeys -nocerts -info -passin pass:
echo
echo Verifying:
echo + openssl verify -CAfile ca.crt "$commonname.crt"
openssl verify -CAfile ca.crt "$commonname.crt"
